import { AxiosResponse, AxiosError, AxiosRequestConfig, AxiosStatic, AxiosInstance, Method } from 'axios';
import { MagicalClass } from 'js-magic';

type IRequestResponse<T = any> = Promise<AxiosResponse<T>>;
type ErrorHandlingStrategy = 'throwError' | 'reject' | 'silent' | 'defaultResponse';
type RequestError = AxiosError<any>;
interface ErrorHandlerClass {
    process(error?: RequestError): unknown;
}
type ErrorHandlerFunction = (error: RequestError) => any;
interface EndpointConfig extends AxiosRequestConfig {
    cancellable?: boolean;
    rejectCancelled?: boolean;
    strategy?: ErrorHandlingStrategy;
    onError?: ErrorHandlerFunction | ErrorHandlerClass;
}
interface RequestHandlerConfig extends EndpointConfig {
    axios: AxiosStatic;
    flattenResponse?: boolean;
    defaultResponse?: any;
    logger?: any;
    onError?: ErrorHandlerFunction | ErrorHandlerClass;
}
interface APIHandlerConfig extends RequestHandlerConfig {
    apiUrl: string;
    endpoints: Record<string, any>;
}
interface IRequestData {
    type: string;
    url: string;
    data?: any;
    config: EndpointConfig;
}

declare type APIQueryParams = Record<string, any>;
declare type APIUrlParams = Record<string, any>;
declare type APIRequestConfig = EndpointConfig;
declare type APIResponse = any;
declare type Endpoint<T = APIQueryParams, T2 = APIUrlParams, T3 = APIResponse> = (queryParams?: T | null, urlParams?: T2, requestConfig?: EndpointConfig) => Promise<T3>;
interface Endpoints {
    [x: string]: Endpoint;
}

/**
 * Generic Request Handler
 * It creates an Axios instance and handles requests within that instance
 * It handles errors depending on a chosen error handling strategy
 */
declare class RequestHandler implements MagicalClass {
    /**
     * @var requestInstance Provider's instance
     */
    requestInstance: AxiosInstance;
    /**
     * @var timeout Request timeout
     */
    timeout: number;
    /**
     * @var cancellable Response cancellation
     */
    cancellable: boolean;
    /**
     * @var strategy Request timeout
     */
    strategy: ErrorHandlingStrategy;
    /**
     * @var flattenResponse Response flattening
     */
    flattenResponse: boolean;
    /**
     * @var defaultResponse Response flattening
     */
    defaultResponse: any;
    /**
     * @var axios Axios instance
     */
    protected axios: AxiosStatic;
    /**
     * @var logger Logger
     */
    protected logger: any;
    /**
     * @var requestErrorService HTTP error service
     */
    protected requestErrorService: any;
    /**
     * @var requestsQueue    Queue of requests
     */
    protected requestsQueue: Map<string, AbortController>;
    /**
     * Creates an instance of HttpRequestHandler
     *
     * @param {string} config.axios                Axios instance
     * @param {string} config.baseURL              Base URL for all API calls
     * @param {number} config.timeout              Request timeout
     * @param {string} config.strategy             Error Handling Strategy
     * @param {string} config.flattenResponse      Whether to flatten response "data" object within "data" one
     * @param {*} config.logger                    Instance of Logger Class
     * @param {*} config.requestErrorService   Instance of Error Service Class
     */
    constructor({ axios, baseURL, timeout, cancellable, strategy, flattenResponse, defaultResponse, logger, onError, ...config }: RequestHandlerConfig);
    /**
     * Get Provider Instance
     *
     * @returns {AxiosInstance} Provider's instance
     */
    getInstance(): AxiosInstance;
    /**
     * Maps all API requests
     *
     * @param {string} url                  Url
     * @param {*} data                      Payload
     * @param {EndpointConfig} config       Config
     * @throws {RequestError}               If request fails
     * @returns {Promise}                   Request response or error info
     */
    __get(prop: string): any;
    /**
     * Prepare Request
     *
     * @param {string} url                  Url
     * @param {*} data                      Payload
     * @param {EndpointConfig} config       Config
     * @throws {RequestError}               If request fails
     * @returns {Promise}                   Request response or error info
     */
    prepareRequest(type: Method, url: string, data?: any, config?: EndpointConfig): Promise<IRequestResponse>;
    /**
     * Build request configuration
     *
     * @param {string} method               Request method
     * @param {string} url                  Request url
     * @param {*}      data                 Request data
     * @param {EndpointConfig} config       Request config
     * @returns {AxiosInstance}             Provider's instance
     */
    protected buildRequestConfig(method: string, url: string, data: any, config: EndpointConfig): EndpointConfig;
    /**
     * Process global Request Error
     *
     * @param {RequestError} error      Error instance
     * @param {EndpointConfig} requestConfig   Per endpoint request config
     * @returns {AxiosInstance} Provider's instance
     */
    protected processRequestError(error: RequestError, requestConfig: EndpointConfig): void;
    /**
     * Output error response depending on chosen strategy
     *
     * @param {RequestError} error      Error instance
     * @param {EndpointConfig} requestConfig   Per endpoint request config
     * @returns {AxiosInstance} Provider's instance
     */
    protected outputErrorResponse(error: RequestError, requestConfig: EndpointConfig): Promise<IRequestResponse>;
    /**
     * Output error response depending on chosen strategy
     *
     * @param {RequestError} error                     Error instance
     * @param {EndpointConfig} _requestConfig    Per endpoint request config
     * @returns {*}                             Error response
     */
    isRequestCancelled(error: RequestError, _requestConfig: EndpointConfig): boolean;
    /**
     * Automatically Cancel Previous Requests
     *
     * @param {EndpointConfig} requestConfig   Per endpoint request config
     * @returns {AxiosInstance} Provider's instance
     */
    protected addCancellationToken(requestConfig: EndpointConfig): {
        signal?: undefined;
    } | {
        signal: AbortSignal;
    };
    /**
     * Handle Request depending on used strategy
     *
     * @param {object} payload                      Payload
     * @param {string} payload.type                 Request type
     * @param {string} payload.url                  Request url
     * @param {*} payload.data                      Request data
     * @param {EndpointConfig} payload.config       Request config
     * @throws {RequestError}
     * @returns {Promise} Response Data
     */
    protected handleRequest({ type, url, data, config, }: IRequestData): Promise<IRequestResponse>;
    /**
     * Process request response
     *
     * @param response Response object
     * @returns {*} Response data
     */
    protected processResponseData(response: any): any;
}

/**
 * Handles dispatching of API requests
 */
declare class ApiHandler implements MagicalClass {
    /**
     * TS Index signature
     */
    [x: string]: any;
    /**
     * @var requestHandler Request Wrapper Instance
     */
    requestHandler: RequestHandler;
    /**
     * Endpoints
     */
    protected endpoints: Record<string, EndpointConfig>;
    /**
     * Logger
     */
    protected logger: any;
    /**
     * Creates an instance of API Handler
     *
     * @param {string} config.apiUrl               Base URL for all API calls
     * @param {number} config.timeout              Request timeout
     * @param {string} config.strategy             Error Handling Strategy
     * @param {string} config.flattenResponse      Whether to flatten response "data" object within "data" one
     * @param {*} config.logger                    Instance of Logger Class
     * @param {*} config.onError                   Instance of Error Service Class
     */
    constructor({ axios, apiUrl, endpoints, timeout, cancellable, strategy, flattenResponse, defaultResponse, logger, onError, ...config }: APIHandlerConfig);
    /**
     * Get Provider Instance
     *
     * @returns {AxiosInstance} Provider's instance
     */
    getInstance(): AxiosInstance;
    /**
     * Maps all API requests
     *
     * @param {*} prop          Caller
     * @returns {Function}      Tailored request function
     */
    __get(prop: any): any;
    /**
     * Handle Single API Request
     *
     * @param {*} args      Arguments
     * @returns {Promise}   Resolvable API provider promise
     */
    handleRequest(...args: any): Promise<IRequestResponse>;
    /**
     * Triggered when trying to use non-existent endpoints
     *
     * @param prop Method Name
     * @returns {Promise}
     */
    protected handleNonImplemented(prop: string): Promise<any>;
}
declare const createApiFetcher: (options: APIHandlerConfig) => ApiHandler;

declare class RequestErrorHandler {
    /**
     * Logger Class
     *
     * @type {*}
     * @memberof RequestErrorHandler
     */
    protected logger: any;
    /**
     * Error Service Class
     *
     * @type {*}
     * @memberof RequestErrorHandler
     */
    requestErrorService: any;
    constructor(logger: any, requestErrorService: any);
    /**
     * Process and Error
     *
     * @param {*} error Error instance or message
     * @throws          Request error context
     * @returns {void}
     */
    process(error: string | Error): void;
}

export { APIHandlerConfig, APIQueryParams, APIRequestConfig, APIResponse, APIUrlParams, ApiHandler, Endpoint, EndpointConfig, Endpoints, ErrorHandlingStrategy, IRequestData, IRequestResponse, RequestError, RequestErrorHandler, RequestHandler, RequestHandlerConfig, createApiFetcher };
